@@ -124,61 +124,62 @@ handleSyncAlarm (CompWindow *w)
 static void
 moveInputFocusToOtherWindow (CompWindow *w)
 {
-    CompDisplay *display = w->screen->display;
+    CompScreen  *s = w->screen;
+    CompDisplay *d = s->display;
+    Bool        focusedAny = FALSE;
 
-    if (w->id == display->activeWindow)
+    if (w->id != d->activeWindow && w->id != d->nextActiveWindow)
+	if (d->activeWindow != None)
+	    return;
+
+    if (w->transientFor && w->transientFor != s->root)
     {
-	CompWindow *ancestor;
-
-	if (w->transientFor && w->transientFor != w->screen->root)
+	CompWindow *ancestor = findWindowAtDisplay (d, w->transientFor);
+	if (ancestor && !(ancestor->type & (CompWindowTypeDesktopMask |
+					    CompWindowTypeDockMask)))
 	{
-	    ancestor = findWindowAtDisplay (display, w->transientFor);
-	    if (ancestor && !(ancestor->type & (CompWindowTypeDesktopMask |
-						CompWindowTypeDockMask)))
-	    {
-		moveInputFocusToWindow (ancestor);
-	    }
-	    else
-		focusDefaultWindow (w->screen);
+	    moveInputFocusToWindow (ancestor);
+	    focusedAny = TRUE;
 	}
-	else if (w->type & (CompWindowTypeDialogMask |
-			    CompWindowTypeModalDialogMask))
-	{
-	    CompWindow *a, *focus = NULL;
-
-	    for (a = w->screen->reverseWindows; a; a = a->prev)
-	    {
-		if (a->clientLeader == w->clientLeader)
-		{
-		    if ((*w->screen->focusWindow) (a))
-		    {
-			if (focus)
-			{
-			    if (a->type & (CompWindowTypeNormalMask |
-					   CompWindowTypeDialogMask |
-					   CompWindowTypeModalDialogMask))
-			    {
-				if (compareWindowActiveness (focus, a) < 0)
-				    focus = a;
-			    }
-			}
-			else
-			    focus = a;
-		    }
-		}
-	    }
-
-	    if (focus && !(focus->type & (CompWindowTypeDesktopMask |
-					  CompWindowTypeDockMask)))
-	    {
-		moveInputFocusToWindow (focus);
-	    }
-	    else
-		focusDefaultWindow (w->screen);
-	}
-	else
-	    focusDefaultWindow (w->screen);
     }
+    else if (w->type & (CompWindowTypeDialogMask |
+			CompWindowTypeModalDialogMask))
+    {
+	CompWindow *a, *focus = NULL;
+
+	for (a = s->reverseWindows; a; a = a->prev)
+	{
+	    if (a->clientLeader != w->clientLeader)
+		continue;
+
+	    if (!(*s->focusWindow) (a))
+		continue;
+
+	    if (!focus)
+	    {
+		focus = a;
+		continue;
+	    }
+
+	    if (a->type & (CompWindowTypeNormalMask |
+			   CompWindowTypeDialogMask |
+			   CompWindowTypeModalDialogMask))
+	    {
+		if (compareWindowActiveness (focus, a) < 0)
+		    focus = a;
+	    }
+	}
+
+	if (focus && !(focus->type & (CompWindowTypeDesktopMask |
+				      CompWindowTypeDockMask)))
+	{
+	    moveInputFocusToWindow (focus);
+	    focusedAny = TRUE;
+	}
+    }
+
+    if (!focusedAny)
+	focusDefaultWindow (s);
 }
 
 static Bool
@@ -1380,7 +1381,9 @@ handleEvent (CompDisplay *d,
 		if (!w->attrib.override_redirect)
 		    setWmState (d, WithdrawnState, w->id);
 
-		w->placed  = FALSE;
+		w->placed     = FALSE;
+		w->unmanaging = w->managed;
+		w->managed    = FALSE;
 	    }
 
 	    unmapWindow (w);
@@ -1641,6 +1644,8 @@ handleEvent (CompDisplay *d,
 	    w = findWindowAtDisplay (d, event->xproperty.window);
 	    if (w)
 	    {
+		s = w->screen;
+
 		if (w->startupId)
 		    free (w->startupId);
 		
@@ -1650,11 +1655,10 @@ handleEvent (CompDisplay *d,
 		{
 		    Time            timestamp = 0;
 		    int             vx, vy, x, y;
-		    CompScreen      *s = w->screen;
 		    CompFocusResult focus;
 
 		    w->initialTimestampSet = FALSE;
-		    applyStartupProperties (w->screen, w);
+		    applyStartupProperties (s, w);
 
 		    if (w->initialTimestampSet)
 			timestamp = w->initialTimestamp;
@@ -1674,7 +1678,7 @@ handleEvent (CompDisplay *d,
 					      timestamp);
 
 		    if (focus == CompFocusAllowed)
-			(*w->screen->activateWindow) (w);
+			(*s->activateWindow) (w);
 		}
 	    }
 	}
@@ -1690,7 +1694,7 @@ handleEvent (CompDisplay *d,
     case ClientMessage:
 	if (event->xclient.message_type == d->winActiveAtom)
 	{
-	    w = findWindowAtDisplay (d, event->xclient.window);
+	    w = findTopLevelWindowAtDisplay (d, event->xclient.window);
 	    if (w)
 	    {
 		CompFocusResult focus = CompFocusAllowed;
@@ -1946,20 +1950,6 @@ handleEvent (CompDisplay *d,
 		}
 	    }
 	}
-	else if (event->xclient.message_type == d->numberOfDesktopsAtom)
-	{
-	    s = findScreenAtDisplay (d, event->xclient.window);
-	    if (s)
-	    {
-		CompOptionValue value;
-
-		value.i = event->xclient.data.l[0];
-
-		(*core.setOptionForPlugin) (&s->base,
-					    "core", "number_of_desktops",
-					    &value);
-	    }
-	}
 	else if (event->xclient.message_type == d->currentDesktopAtom)
 	{
 	    s = findScreenAtDisplay (d, event->xclient.window);
@@ -2032,6 +2022,7 @@ handleEvent (CompDisplay *d,
 	    {
 		CompFocusResult        focus;
 		CompStackingUpdateMode stackingMode;
+		Bool                   initiallyMinimized;
 
 		if (!w->placed)
 		{
@@ -2040,11 +2031,11 @@ handleEvent (CompDisplay *d,
 		    XWindowChanges xwc;
 		    unsigned int   xwcm, source;
 
-		    /* adjust for gravity */
+		    /* adjust for gravity, but only for frame size */
 		    xwc.x      = w->serverX;
 		    xwc.y      = w->serverY;
-		    xwc.width  = w->serverWidth;
-		    xwc.height = w->serverHeight;
+		    xwc.width  = 0;
+		    xwc.height = 0;
 
 		    xwcm = adjustConfigureRequestForGravity (w, &xwc,
 							     CWX | CWY,
@@ -2071,26 +2062,38 @@ handleEvent (CompDisplay *d,
 		focus = allowWindowFocus (w, NO_FOCUS_MASK,
 					  w->screen->x, w->screen->y, 0);
 
-		if (focus == CompFocusDenied)
+		if ((focus == CompFocusDenied || focus == CompFocusPrevent) && (w->type & ~NO_FOCUS_MASK))
 		    stackingMode = CompStackingUpdateModeInitialMapDeniedFocus;
 		else
 		    stackingMode = CompStackingUpdateModeInitialMap;
 
 		updateWindowAttributes (w, stackingMode);
 
-		if (w->minimized)
+		initiallyMinimized = w->hints &&
+		                     w->hints->initial_state == IconicState;
+
+		if (w->minimized && !initiallyMinimized)
 		    unminimizeWindow (w);
 
 		(*w->screen->leaveShowDesktopMode) (w->screen, w);
 
-		if (focus == CompFocusAllowed && !onCurrentDesktop (w))
-		    setCurrentDesktop (w->screen, w->desktop);
+		if (!initiallyMinimized)
+		{
+		    if (focus == CompFocusAllowed && !onCurrentDesktop (w))
+			setCurrentDesktop (w->screen, w->desktop);
 
-		if (!(w->state & CompWindowStateHiddenMask))
-		    showWindow (w);
+		    if (!(w->state & CompWindowStateHiddenMask))
+			showWindow (w);
 
-		if (focus == CompFocusAllowed)
-		    moveInputFocusToWindow (w);
+		    if (focus == CompFocusAllowed)
+			moveInputFocusToWindow (w);
+		}
+		else
+		{
+		    minimizeWindow (w);
+		    changeWindowState (w, w->state | CompWindowStateHiddenMask);
+		    updateClientListForScreen (w->screen);
+		}
 	    }
 
 	    setWindowProp (d, w->id, d->winDesktopAtom, w->desktop);
@@ -2207,6 +2210,25 @@ handleEvent (CompDisplay *d,
 		state &= ~CompWindowStateDemandsAttentionMask;
 		changeWindowState (w, state);
 	    }
+	    else
+	    {
+		d->activeWindow = None;
+
+		s = findScreenAtDisplay (d, event->xfocus.window);
+		if (s)
+		{
+		    if (event->xfocus.detail == NotifyDetailNone ||
+			(event->xfocus.mode == NotifyNormal &&
+			 event->xfocus.detail == NotifyInferior))
+		    {
+			/* we don't want the root window to get focus */
+			focusDefaultWindow (s);
+		    }
+		}
+	    }
+
+	    if (d->nextActiveWindow == event->xfocus.window)
+		d->nextActiveWindow = None;
 	}
 	break;
     case EnterNotify:
@@ -2292,32 +2314,50 @@ handleEvent (CompDisplay *d,
 
 	    if (w)
 	    {
+		XserverRegion parts = XFixesCreateRegion(de->display, NULL, 0);
+		XRectangle *rects;
+		int nRects;
+
 		w->texture->oldMipmaps = TRUE;
 
+		// Get the damage region
+		XDamageSubtract(de->display, de->damage, None, parts);
+		rects = XFixesFetchRegion(de->display, parts, &nRects);
+		XFixesDestroyRegion(de->display, parts);
+
 		if (w->syncWait)
 		{
-		    if (w->nDamage == w->sizeDamage)
+		    int i;
+
+		    if (w->nDamage + nRects - 1 >= w->sizeDamage)
 		    {
 			w->damageRects = realloc (w->damageRects,
-						  (w->sizeDamage + 1) *
+						  (w->sizeDamage + nRects) *
 						  sizeof (XRectangle));
-			w->sizeDamage += 1;
+			w->sizeDamage += nRects;
 		    }
 
-		    w->damageRects[w->nDamage].x      = de->area.x;
-		    w->damageRects[w->nDamage].y      = de->area.y;
-		    w->damageRects[w->nDamage].width  = de->area.width;
-		    w->damageRects[w->nDamage].height = de->area.height;
-		    w->nDamage++;
+		    for (i = 0; i < nRects; i++)
+		    {
+			w->damageRects[w->nDamage] = rects[i];
+			w->nDamage++;
+		    }
 		}
 		else
 		{
-		    handleWindowDamageRect (w,
-					    de->area.x,
-					    de->area.y,
-					    de->area.width,
-					    de->area.height);
+		    int i;
+
+		    for (i = 0; i < nRects; i++)
+		    {
+			handleWindowDamageRect (w,
+						rects[i].x,
+						rects[i].y,
+						rects[i].width,
+						rects[i].height);
+		    }
 		}
+
+		XFree(rects);
 	    }
 	}
 	else if (d->shapeExtension &&
