@@ -43,6 +43,9 @@
 #include <X11/Xproto.h>
 #include <X11/extensions/Xrandr.h>
 #include <X11/extensions/shape.h>
+#ifdef HAVE_XINPUT2
+#include <X11/extensions/XInput2.h>
+#endif
 #include <X11/cursorfont.h>
 
 #include <compiz-core.h>
@@ -255,6 +258,72 @@ setVirtualScreenSize (CompScreen *screen,
 		      int	 hsize,
 		      int	 vsize)
 {
+    /* if hsize or vsize is being reduced */
+    if (hsize < screen->hsize ||
+	vsize < screen->vsize)
+    {
+	CompWindow *w;
+	int        tx = 0;
+	int        ty = 0;
+
+	if (screen->x >= hsize)
+	    tx = screen->x - (hsize - 1);
+	if (screen->y >= vsize)
+	    ty = screen->y - (vsize - 1);
+
+	if (tx != 0 || ty != 0)
+	    moveScreenViewport (screen, tx, ty, TRUE);
+
+	/* Move windows that were in one of the deleted viewports into the
+	   closest viewport */
+	for (w = screen->windows; w; w = w->next)
+	{
+	    int moveX = 0;
+	    int moveY = 0;
+
+	    if (windowOnAllViewports (w))
+		continue;
+
+	    /* Find which viewport the (inner) window's top-left corner falls
+	       in, and check if it's outside the new viewport horizontal and
+	       vertical index range */
+	    if (hsize < screen->hsize)
+	    {
+		int vpX;   /* x index of a window's vp */
+
+		vpX = w->serverX / screen->width;
+		if (w->serverX < 0)
+		    vpX -= 1;
+
+		vpX += screen->x; /* Convert relative to absolute vp index */
+
+		/* Move windows too far right to left */
+		if (vpX >= hsize)
+		    moveX = ((hsize - 1) - vpX) * screen->width;
+	    }
+	    if (vsize < screen->vsize)
+	    {
+		int vpY;   /* y index of a window's vp */
+
+		vpY = w->serverY / screen->height;
+		if (w->serverY < 0)
+		    vpY -= 1;
+
+		vpY += screen->y; /* Convert relative to absolute vp index */
+
+		/* Move windows too far right to left */
+		if (vpY >= vsize)
+		    moveY = ((vsize - 1) - vpY) * screen->height;
+	    }
+
+	    if (moveX != 0 || moveY != 0)
+	    {
+		moveWindow (w, moveX, moveY, TRUE, TRUE);
+		syncWindowPosition (w);
+	    }
+	}
+    }
+
     screen->hsize = hsize;
     screen->vsize = vsize;
 
@@ -588,13 +657,6 @@ setScreenOption (CompPlugin	 *plugin,
 	    return TRUE;
 	}
 	break;
-    case COMP_SCREEN_OPTION_NUMBER_OF_DESKTOPS:
-	if (compSetIntOption (o, value))
-	{
-	    setNumberOfDesktops (screen, o->value.i);
-	    return TRUE;
-	}
-	break;
     case COMP_SCREEN_OPTION_DEFAULT_ICON:
 	if (compSetStringOption (o, value))
 	    return updateDefaultIcon (screen);
@@ -635,7 +697,7 @@ const CompMetadataOptionInfo coreScreenOptionInfo[COMP_SCREEN_OPTION_NUM] = {
     { "unredirect_fullscreen_windows", "bool", 0, 0, 0 },
     { "default_icon", "string", 0, 0, 0 },
     { "sync_to_vblank", "bool", 0, 0, 0 },
-    { "number_of_desktops", "int", "<min>1</min>", 0, 0 },
+    { "__padding__", "bool", 0, 0, 0 },
     { "detect_outputs", "bool", 0, 0, 0 },
     { "outputs", "list", "<type>string</type>", 0, 0 },
     { "overlapping_outputs", "int",
@@ -650,6 +712,23 @@ const CompMetadataOptionInfo coreScreenOptionInfo[COMP_SCREEN_OPTION_NUM] = {
 static void
 updateStartupFeedback (CompScreen *s)
 {
+#ifdef HAVE_XINPUT2
+    int clientPointer;
+
+    XIGetClientPointer (s->display->display, None, &clientPointer);
+
+    if (s->startupSequences)
+    {
+	XIDefineCursor (s->display->display, clientPointer,
+			s->root, s->busyCursor);
+    }
+    else
+    {
+	XIDefineCursor (s->display->display, clientPointer,
+			s->root, s->normalCursor);
+    }
+#endif
+
     if (s->startupSequences)
 	XDefineCursor (s->display->display, s->root, s->busyCursor);
     else
@@ -704,9 +783,10 @@ addSequence (CompScreen        *screen,
     screen->startupSequences = s;
 
     if (!screen->startupSequenceTimeoutHandle)
-	compAddTimeout (1000, 1500,
-			startupSequenceTimeout,
-			screen);
+	screen->startupSequenceTimeoutHandle =
+	    compAddTimeout (1000, 1500,
+			    startupSequenceTimeout,
+			    screen);
 
     updateStartupFeedback (screen);
 }
@@ -746,6 +826,28 @@ removeSequence (CompScreen        *screen,
     updateStartupFeedback (screen);
 }
 
+static void
+removeAllSequences (CompScreen *screen)
+{
+    CompStartupSequence *s;
+    CompStartupSequence *sNext;
+
+    for (s = screen->startupSequences; s; s = sNext)
+    {
+	sNext = s->next;
+	sn_startup_sequence_unref (s->sequence);
+	free (s);
+    }
+    screen->startupSequences = NULL;
+
+    if (screen->startupSequenceTimeoutHandle)
+    {
+	compRemoveTimeout (screen->startupSequenceTimeoutHandle);
+	screen->startupSequenceTimeoutHandle = 0;
+    }
+    updateStartupFeedback (screen);
+}
+
 static void
 compScreenSnEvent (SnMonitorEvent *event,
 		   void           *userData)
@@ -888,8 +990,7 @@ reshape (CompScreen *s,
     perspective (s->projection, 60.0f, 1.0f, 0.1f, 100.0f);
 
     glMatrixMode (GL_PROJECTION);
-    glLoadIdentity ();
-    glMultMatrixf (s->projection);
+    glLoadMatrixf (s->projection);
     glMatrixMode (GL_MODELVIEW);
 
     s->region.rects = &s->region.extents;
@@ -913,6 +1014,12 @@ reshape (CompScreen *s,
     s->fullscreenOutput.workArea.width   = w;
     s->fullscreenOutput.workArea.height  = h;
 
+    if (w > s->maxTextureSize || h > s->maxTextureSize)
+	compLogMessage ("core", CompLogLevelWarn,
+	                "Screen size %d x %d exceeds maximum texture size (%d),"
+			" expect rendering problems of e.g. desktop windows.",
+			w, h, s->maxTextureSize);
+
     updateScreenEdges (s);
 }
 
@@ -1110,106 +1217,114 @@ setSupportingWmCheck (CompScreen *s)
 		     (unsigned char *) &s->grabWindow, 1);
 }
 
-static void
-setSupported (CompScreen *s)
+static unsigned int
+addSupportedAtoms (CompScreen   *s,
+		   Atom         *atoms,
+		   unsigned int size)
 {
-    CompDisplay *d = s->display;
-    Atom	data[256];
-    int		i = 0;
+    CompDisplay  *d = s->display;
+    unsigned int count = 0;
 
-    data[i++] = d->supportedAtom;
-    data[i++] = d->supportingWmCheckAtom;
+    atoms[count++] = d->utf8StringAtom;
 
-    data[i++] = d->utf8StringAtom;
+    atoms[count++] = d->clientListAtom;
+    atoms[count++] = d->clientListStackingAtom;
 
-    data[i++] = d->clientListAtom;
-    data[i++] = d->clientListStackingAtom;
+    atoms[count++] = d->winActiveAtom;
 
-    data[i++] = d->winActiveAtom;
+    atoms[count++] = d->desktopViewportAtom;
+    atoms[count++] = d->desktopGeometryAtom;
+    atoms[count++] = d->currentDesktopAtom;
+    atoms[count++] = d->numberOfDesktopsAtom;
+    atoms[count++] = d->showingDesktopAtom;
 
-    data[i++] = d->desktopViewportAtom;
-    data[i++] = d->desktopGeometryAtom;
-    data[i++] = d->currentDesktopAtom;
-    data[i++] = d->numberOfDesktopsAtom;
-    data[i++] = d->showingDesktopAtom;
+    atoms[count++] = d->workareaAtom;
 
-    data[i++] = d->workareaAtom;
+    atoms[count++] = d->wmNameAtom;
 
-    data[i++] = d->wmNameAtom;
-/*
-    data[i++] = d->wmVisibleNameAtom;
-*/
+    atoms[count++] = d->wmStrutAtom;
+    atoms[count++] = d->wmStrutPartialAtom;
 
-    data[i++] = d->wmStrutAtom;
-    data[i++] = d->wmStrutPartialAtom;
+    atoms[count++] = d->wmUserTimeAtom;
+    atoms[count++] = d->frameExtentsAtom;
+    atoms[count++] = d->frameWindowAtom;
 
-/*
-    data[i++] = d->wmPidAtom;
-*/
+    atoms[count++] = d->winStateAtom;
+    atoms[count++] = d->winStateModalAtom;
+    atoms[count++] = d->winStateStickyAtom;
+    atoms[count++] = d->winStateMaximizedVertAtom;
+    atoms[count++] = d->winStateMaximizedHorzAtom;
+    atoms[count++] = d->winStateShadedAtom;
+    atoms[count++] = d->winStateSkipTaskbarAtom;
+    atoms[count++] = d->winStateSkipPagerAtom;
+    atoms[count++] = d->winStateHiddenAtom;
+    atoms[count++] = d->winStateFullscreenAtom;
+    atoms[count++] = d->winStateAboveAtom;
+    atoms[count++] = d->winStateBelowAtom;
+    atoms[count++] = d->winStateDemandsAttentionAtom;
 
-    data[i++] = d->wmUserTimeAtom;
-    data[i++] = d->frameExtentsAtom;
-    data[i++] = d->frameWindowAtom;
-
-    data[i++] = d->winStateAtom;
-    data[i++] = d->winStateModalAtom;
-    data[i++] = d->winStateStickyAtom;
-    data[i++] = d->winStateMaximizedVertAtom;
-    data[i++] = d->winStateMaximizedHorzAtom;
-    data[i++] = d->winStateShadedAtom;
-    data[i++] = d->winStateSkipTaskbarAtom;
-    data[i++] = d->winStateSkipPagerAtom;
-    data[i++] = d->winStateHiddenAtom;
-    data[i++] = d->winStateFullscreenAtom;
-    data[i++] = d->winStateAboveAtom;
-    data[i++] = d->winStateBelowAtom;
-    data[i++] = d->winStateDemandsAttentionAtom;
-
-    data[i++] = d->winOpacityAtom;
-    data[i++] = d->winBrightnessAtom;
+    atoms[count++] = d->winOpacityAtom;
+    atoms[count++] = d->winBrightnessAtom;
 
     if (s->canDoSaturated)
     {
-	data[i++] = d->winSaturationAtom;
-	data[i++] = d->winStateDisplayModalAtom;
+	atoms[count++] = d->winSaturationAtom;
+	atoms[count++] = d->winStateDisplayModalAtom;
     }
 
-    data[i++] = d->wmAllowedActionsAtom;
+    atoms[count++] = d->wmAllowedActionsAtom;
 
-    data[i++] = d->winActionMoveAtom;
-    data[i++] = d->winActionResizeAtom;
-    data[i++] = d->winActionStickAtom;
-    data[i++] = d->winActionMinimizeAtom;
-    data[i++] = d->winActionMaximizeHorzAtom;
-    data[i++] = d->winActionMaximizeVertAtom;
-    data[i++] = d->winActionFullscreenAtom;
-    data[i++] = d->winActionCloseAtom;
-    data[i++] = d->winActionShadeAtom;
-    data[i++] = d->winActionChangeDesktopAtom;
-    data[i++] = d->winActionAboveAtom;
-    data[i++] = d->winActionBelowAtom;
+    atoms[count++] = d->winActionMoveAtom;
+    atoms[count++] = d->winActionResizeAtom;
+    atoms[count++] = d->winActionStickAtom;
+    atoms[count++] = d->winActionMinimizeAtom;
+    atoms[count++] = d->winActionMaximizeHorzAtom;
+    atoms[count++] = d->winActionMaximizeVertAtom;
+    atoms[count++] = d->winActionFullscreenAtom;
+    atoms[count++] = d->winActionCloseAtom;
+    atoms[count++] = d->winActionShadeAtom;
+    atoms[count++] = d->winActionChangeDesktopAtom;
+    atoms[count++] = d->winActionAboveAtom;
+    atoms[count++] = d->winActionBelowAtom;
 
-    data[i++] = d->winTypeAtom;
-    data[i++] = d->winTypeDesktopAtom;
-    data[i++] = d->winTypeDockAtom;
-    data[i++] = d->winTypeToolbarAtom;
-    data[i++] = d->winTypeMenuAtom;
-    data[i++] = d->winTypeSplashAtom;
-    data[i++] = d->winTypeDialogAtom;
-    data[i++] = d->winTypeUtilAtom;
-    data[i++] = d->winTypeNormalAtom;
+    atoms[count++] = d->winTypeAtom;
+    atoms[count++] = d->winTypeDesktopAtom;
+    atoms[count++] = d->winTypeDockAtom;
+    atoms[count++] = d->winTypeToolbarAtom;
+    atoms[count++] = d->winTypeMenuAtom;
+    atoms[count++] = d->winTypeSplashAtom;
+    atoms[count++] = d->winTypeDialogAtom;
+    atoms[count++] = d->winTypeUtilAtom;
+    atoms[count++] = d->winTypeNormalAtom;
 
-    data[i++] = d->wmDeleteWindowAtom;
-    data[i++] = d->wmPingAtom;
+    atoms[count++] = d->wmDeleteWindowAtom;
+    atoms[count++] = d->wmPingAtom;
 
-    data[i++] = d->wmMoveResizeAtom;
-    data[i++] = d->moveResizeWindowAtom;
-    data[i++] = d->restackWindowAtom;
+    atoms[count++] = d->wmMoveResizeAtom;
+    atoms[count++] = d->moveResizeWindowAtom;
+    atoms[count++] = d->restackWindowAtom;
 
-    data[i++] = d->wmFullscreenMonitorsAtom;
+    atoms[count++] = d->wmFullscreenMonitorsAtom;
+
+    assert (count < size);
+
+    return count;
+}
+
+void
+setSupportedWmHints (CompScreen *s)
+{
+    CompDisplay  *d = s->display;
+    Atom	 data[256];
+    unsigned int count = 0;
+
+    data[count++] = d->supportedAtom;
+    data[count++] = d->supportingWmCheckAtom;
+
+    count += (*s->addSupportedAtoms) (s, data + count, 256 - count);
 
     XChangeProperty (d->display, s->root, d->supportedAtom, XA_ATOM, 32,
-		     PropModeReplace, (unsigned char *) data, i);
+		     PropModeReplace, (unsigned char *) data, count);
 }
 
 static void
@@ -1625,7 +1740,10 @@ addScreen (CompDisplay *display,
     Window		 *children;
     unsigned int	 nchildren;
     int			 defaultDepth, nvisinfo, nElements, value, i;
-    const char		 *glxExtensions, *glExtensions;
+#ifdef HAVE_XINPUT2
+    int clientPointer;
+#endif
+    const char		 *glxExtensions, *glExtensions, *glRenderer;
     XSetWindowAttributes attrib;
     GLfloat		 globalAmbient[]  = { 0.1f, 0.1f,  0.1f, 0.1f };
     GLfloat		 ambientLight[]   = { 0.0f, 0.0f,  0.0f, 0.0f };
@@ -1802,6 +1920,7 @@ addScreen (CompDisplay *display,
     s->windowStateChangeNotify = windowStateChangeNotify;
 
     s->outputChangeNotify = outputChangeNotify;
+    s->addSupportedAtoms  = addSupportedAtoms;
 
     s->initWindowWalker = initWindowWalker;
 
@@ -1891,6 +2010,34 @@ addScreen (CompDisplay *display,
     glxExtensions = glXQueryExtensionsString (dpy, screenNum);
     if (!strstr (glxExtensions, "GLX_EXT_texture_from_pixmap"))
     {
+	/* try again with indirect rendering */
+	if (!indirectRendering)
+	{
+	    char **copy;
+
+	    copy = (char **)malloc ((programArgc + 2) * sizeof (char *));
+	    for (i = 0; i < programArgc; i++)
+	    {
+		copy[i] = strdup (programArgv[i]);
+	    }
+	    copy[i++] = "--indirect-rendering";
+	    copy[i] = NULL;
+	    execvp (programName, copy);
+
+	    /* if we made it here execvp failed */
+	    for (i = 0; copy[i] != NULL; i++)
+	    {
+		free (copy[i]);
+	    }
+	    free (copy);
+
+	    compLogMessage ("core", CompLogLevelFatal,
+			    "Failed to launch with --indirect-rendering");
+	    XFree (visinfo);
+
+	    return FALSE;
+	}
+
 	compLogMessage ("core", CompLogLevelFatal,
 			"GLX_EXT_texture_from_pixmap is missing");
 	XFree (visinfo);
@@ -1921,6 +2068,8 @@ addScreen (CompDisplay *display,
 	getProcAddress (s, "glXGetFBConfigAttrib");
     s->createPixmap = (GLXCreatePixmapProc)
 	getProcAddress (s, "glXCreatePixmap");
+    s->destroyPixmap = (GLXDestroyPixmapProc)
+	getProcAddress (s, "glXDestroyPixmap");
 
     if (!s->bindTexImage)
     {
@@ -1939,7 +2088,8 @@ addScreen (CompDisplay *display,
     if (!s->queryDrawable     ||
 	!s->getFBConfigs      ||
 	!s->getFBConfigAttrib ||
-	!s->createPixmap)
+	!s->createPixmap      ||
+	!s->destroyPixmap)
     {
 	compLogMessage ("core", CompLogLevelFatal,
 			"fbconfig functions missing");
@@ -1973,6 +2123,19 @@ addScreen (CompDisplay *display,
 	return FALSE;
     }
 
+    if (!getenv ("SKIP_CHECKS") || strcmp (getenv ("SKIP_CHECKS"), "yes") != 0)
+    {
+	glRenderer = (const char *) glGetString (GL_RENDERER);
+	if (glRenderer != NULL &&
+	(strcmp (glRenderer, "Software Rasterizer") == 0 ||
+		strcmp (glRenderer, "Mesa X11") == 0))
+	{
+	    compLogMessage ("core", CompLogLevelFatal,
+	                    "Software rendering detected.");
+	    return FALSE;
+        }
+    }
+
     s->textureNonPowerOfTwo = 0;
     if (strstr (glExtensions, "GL_ARB_texture_non_power_of_two"))
 	s->textureNonPowerOfTwo = 1;
@@ -2020,6 +2183,10 @@ addScreen (CompDisplay *display,
 	strstr (glExtensions, "GL_SGIS_texture_border_clamp"))
 	s->textureBorderClamp = 1;
 
+    s->activeTexture       = NULL;
+    s->clientActiveTexture = NULL;
+    s->multiTexCoord2f     = NULL;
+
     s->maxTextureUnits = 1;
     if (strstr (glExtensions, "GL_ARB_multitexture"))
     {
@@ -2034,6 +2201,14 @@ addScreen (CompDisplay *display,
 	    glGetIntegerv (GL_MAX_TEXTURE_UNITS_ARB, &s->maxTextureUnits);
     }
 
+    s->genPrograms             = NULL;
+    s->deletePrograms          = NULL;
+    s->bindProgram             = NULL;
+    s->programString           = NULL;
+    s->programEnvParameter4f   = NULL;
+    s->programLocalParameter4f = NULL;
+    s->getProgramiv            = NULL;
+
     s->fragmentProgram = 0;
     if (strstr (glExtensions, "GL_ARB_fragment_program"))
     {
@@ -2062,8 +2237,15 @@ addScreen (CompDisplay *display,
 	    s->fragmentProgram = 1;
     }
 
+    s->genFramebuffers        = NULL;
+    s->deleteFramebuffers     = NULL;
+    s->bindFramebuffer        = NULL;
+    s->checkFramebufferStatus = NULL;
+    s->framebufferTexture2D   = NULL;
+    s->generateMipmap         = NULL;
+
     s->fbo = 0;
-    if (strstr (glExtensions, "GL_EXT_framebuffer_object"))
+    if (!noFBO && strstr (glExtensions, "GL_EXT_framebuffer_object"))
     {
 	s->genFramebuffers = (GLGenFramebuffersProc)
 	    getProcAddress (s, "glGenFramebuffersEXT");
@@ -2354,12 +2536,18 @@ addScreen (CompDisplay *display,
 
     setDesktopHints (s);
     setSupportingWmCheck (s);
-    setSupported (s);
+    setSupportedWmHints (s);
 
     s->normalCursor = XCreateFontCursor (dpy, XC_left_ptr);
     s->busyCursor   = XCreateFontCursor (dpy, XC_watch);
 
+#ifdef HAVE_XINPUT2
+    XIGetClientPointer (dpy, None, &clientPointer);
+
+    XIDefineCursor (dpy, clientPointer, s->root, s->normalCursor);
+#else
     XDefineCursor (dpy, s->root, s->normalCursor);
+#endif
 
     s->filter[NOTHING_TRANS_FILTER] = COMP_TEXTURE_FILTER_FAST;
     s->filter[SCREEN_TRANS_FILTER]  = COMP_TEXTURE_FILTER_GOOD;
@@ -2384,6 +2572,8 @@ removeScreen (CompScreen *s)
     else
 	d->screens = NULL;
 
+    removeAllSequences (s);
+
     while (s->windows)
 	removeWindow (s->windows);
 
@@ -2475,7 +2665,7 @@ focusDefaultWindow (CompScreen *s)
 	if (w && (*w->screen->focusWindow) (w))
 	{
 	    if (!(w->type & (CompWindowTypeDesktopMask |
-			    CompWindowTypeDockMask)))
+			     CompWindowTypeDockMask)))
 		focus = w;
 	}
 	else
@@ -2496,7 +2686,7 @@ focusDefaultWindow (CompScreen *s)
 	    {
 		w = findTopLevelWindowAtDisplay (d, childReturn);
 
-		if (w && (*w->screen->focusWindow) (w))
+		if (w && (*s->focusWindow) (w))
 		{
 		    if (!(w->type & (CompWindowTypeDesktopMask |
 				     CompWindowTypeDockMask)))
@@ -2513,19 +2703,20 @@ focusDefaultWindow (CompScreen *s)
 	    if (w->type & CompWindowTypeDockMask)
 		continue;
 
-	    if ((*s->focusWindow) (w))
+	    if (!(*s->focusWindow) (w))
+		continue;
+
+	    if (!focus)
 	    {
-		if (focus)
-		{
-		    if (w->type & (CompWindowTypeNormalMask |
-				   CompWindowTypeDialogMask |
-				   CompWindowTypeModalDialogMask))
-		    {
-			if (compareWindowActiveness (focus, w) < 0)
-			    focus = w;
-		    }
-		}
-		else
+		focus = w;
+		continue;
+	    }
+
+	    if (w->type & (CompWindowTypeNormalMask |
+			   CompWindowTypeDialogMask |
+			   CompWindowTypeModalDialogMask))
+	    {
+		if (compareWindowActiveness (focus, w) < 0)
 		    focus = w;
 	    }
 	}
@@ -3914,35 +4105,6 @@ getCurrentOutputExtents (CompScreen *s,
 	*y2 = s->outputDev[s->currentOutputDev].region.extents.y2;
 }
 
-void
-setNumberOfDesktops (CompScreen   *s,
-		     unsigned int nDesktop)
-{
-    CompWindow *w;
-
-    if (nDesktop < 1 || nDesktop >= 0xffffffff)
-	return;
-
-    if (nDesktop == s->nDesktop)
-	return;
-
-    if (s->currentDesktop >= nDesktop)
-	s->currentDesktop = nDesktop - 1;
-
-    for (w = s->windows; w; w = w->next)
-    {
-	if (w->desktop == 0xffffffff)
-	    continue;
-
-	if (w->desktop >= nDesktop)
-	    setDesktopForWindow (w, nDesktop - 1);
-    }
-
-    s->nDesktop = nDesktop;
-
-    setDesktopHints (s);
-}
-
 void
 setCurrentDesktop (CompScreen   *s,
 		   unsigned int desktop)
